# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from pathlib import Path
from subprocess import DEVNULL, PIPE, CalledProcessError, run
from tempfile import TemporaryDirectory
from typing import cast

import deprecation
from packaging.version import Version
from PIL import Image

import pikepdf
from pikepdf._exceptions import DependencyError


@deprecation.deprecated(
    deprecated_in="5.1.5", removed_in="6.0", details="Use extract_jbig2_bytes instead"
)
def extract_jbig2(
    im_obj: pikepdf.Object, globals_obj: pikepdf.Object | None = None
) -> Image.Image:  # pragma: no cover

    with TemporaryDirectory(prefix='pikepdf-', suffix='.jbig2') as tmpdir:
        image_path = Path(tmpdir) / "image"
        global_path = Path(tmpdir) / "global"
        output_path = Path(tmpdir) / "outfile"

        args = [
            "jbig2dec",
            "--embedded",
            "--format",
            "png",
            "--output",
            os.fspath(output_path),
        ]

        # Get the raw stream, because we can't decode im_obj - that is why we are here
        # (Strictly speaking we should remove any non-JBIG2 filters if double encoded)
        image_path.write_bytes(cast(memoryview, im_obj.get_raw_stream_buffer()))

        if globals_obj is not None:
            # For globals, we do want to remove any encoding since it's just a binary
            # blob and won't be marked with /JBIG2Decode
            global_path.write_bytes(cast(memoryview, globals_obj.get_stream_buffer()))
            args.append(os.fspath(global_path))

        args.append(os.fspath(image_path))

        run(args, stdout=DEVNULL, check=True)
        im = Image.open(output_path)
        im.load()  # Load pixel data into memory so file/tempdir can be closed
        return im


def _extract_jbig2_bytes(jbig2: bytes, jbig2_globals: bytes) -> bytes:
    with TemporaryDirectory(prefix='pikepdf-', suffix='.jbig2') as tmpdir:
        image_path = Path(tmpdir) / "image"
        global_path = Path(tmpdir) / "global"
        output_path = Path(tmpdir) / "outfile"

        args = [
            "jbig2dec",
            "--embedded",
            "--format",
            "png",
            "--output",
            os.fspath(output_path),
        ]

        # Get the raw stream, because we can't decode im_obj - that is why we are here
        # (Strictly speaking we should remove any non-JBIG2 filters if double encoded)
        image_path.write_bytes(jbig2)

        if len(jbig2_globals) > 0:
            global_path.write_bytes(jbig2_globals)
            args.append(os.fspath(global_path))

        args.append(os.fspath(image_path))

        run(args, stdout=DEVNULL, check=True)
        with Image.open(output_path) as im:
            return im.tobytes()


@deprecation.deprecated(
    deprecated_in="5.2.0",
    removed_in="6.0",
    details="Use jbig2.get_decoder() interface instead",
)
def extract_jbig2_bytes(
    jbig2: bytes, jbig2_globals: bytes
) -> bytes:  # pragma: no cover
    return _extract_jbig2_bytes(jbig2, jbig2_globals)


def _check_jbig2dec_available() -> None:  # pragma: no cover
    try:
        proc = run(['jbig2dec', '--version'], stdout=PIPE, check=True, encoding='ascii')
    except (CalledProcessError, FileNotFoundError) as e:
        raise DependencyError("jbig2dec - not installed or not found") from e
    else:
        result = proc.stdout
        version_str = result.replace('jbig2dec', '').strip()  # returns "jbig2dec 0.xx"
        version = Version(version_str)
        if version < Version('0.15'):
            raise DependencyError("jbig2dec is too old (older than version 0.15)")


@deprecation.deprecated(
    deprecated_in="5.2.0",
    removed_in="6.0",
    details="Use jbig2.get_decoder() interface instead",
)
def check_jbig2dec_available() -> None:  # pragma: no cover
    _check_jbig2dec_available()


@deprecation.deprecated(
    deprecated_in="5.2.0",
    removed_in="6.0",
    details="Use jbig2.get_decoder() interface instead",
)
def jbig2dec_available() -> bool:  # pragma: no cover
    try:
        _check_jbig2dec_available()
    except (DependencyError, CalledProcessError, FileNotFoundError):
        return False
    else:
        return True


class JBIG2DecoderInterface(ABC):
    """pikepdf's C++ expects this Python interface to be available for JBIG2."""

    @abstractmethod
    def check_available(self) -> None:
        """Check if decoder is available. Throws DependencyError if not."""

    @abstractmethod
    def decode_jbig2(self, jbig2: bytes, jbig2_globals: bytes) -> bytes:
        """Decode JBIG2 from jbig2 and globals, returning decoded bytes."""

    def available(self) -> bool:
        """Returns True if decoder is available."""
        try:
            self.check_available()
        except DependencyError:
            return False
        else:
            return True


class JBIG2Decoder(JBIG2DecoderInterface):
    def check_available(self) -> None:
        version = self._version()
        if version < Version('0.15'):
            raise DependencyError("jbig2dec is too old (older than version 0.15)")

    def decode_jbig2(self, jbig2: bytes, jbig2_globals: bytes) -> bytes:
        return _extract_jbig2_bytes(jbig2, jbig2_globals)

    def _version(self) -> Version:
        try:
            proc = run(
                ['jbig2dec', '--version'], stdout=PIPE, check=True, encoding='ascii'
            )
        except (CalledProcessError, FileNotFoundError) as e:
            raise DependencyError("jbig2dec - not installed or not found") from e
        else:
            result = proc.stdout
            version_str = result.replace(
                'jbig2dec', ''
            ).strip()  # returns "jbig2dec 0.xx"
            return Version(version_str)


_jbig2_decoder = JBIG2Decoder()


def get_decoder():
    return _jbig2_decoder



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations

import sys
from glob import glob
from itertools import chain
from os import environ
from os.path import exists, join
from platform import machine
from typing import cast

from pybind11.setup_helpers import ParallelCompile, Pybind11Extension, build_ext
from setuptools import Extension, setup

extra_includes = []
extra_library_dirs = []
qpdf_source_tree = environ.get('QPDF_SOURCE_TREE', '')

# If CFLAGS is defined, disable any efforts to shim the build, because
# the caller is probably a maintainer and knows what they are doing.
cflags_defined = bool(environ.get('CFLAGS', ''))

if not cflags_defined:
    if qpdf_source_tree:
        # Point this to qpdf source tree built with shared libaries
        extra_includes.append(join(qpdf_source_tree, 'include'))
        extra_library_dirs.append(join(qpdf_source_tree, 'libqpdf/build/.libs'))

    if 'bsd' in sys.platform:
        extra_includes.append('/usr/local/include')
    elif 'darwin' in sys.platform and machine() == 'arm64':
        extra_includes.append('/opt/homebrew/include')
        extra_library_dirs.append('/opt/homebrew/lib')


for extra_path in chain([qpdf_source_tree], extra_includes, extra_library_dirs):
    if extra_path and not exists(extra_path):
        raise FileNotFoundError(extra_path)

# Use cast because mypy has trouble seeing Pybind11Extension is a subclass of
# Extension.
extmodule: Extension = cast(
    Extension,
    Pybind11Extension(
        'pikepdf._qpdf',
        sorted(glob('src/qpdf/*.cpp')),
        depends=sorted(glob('src/qpdf/*.h')),
        include_dirs=[
            # Path to pybind11 headers
            *extra_includes,
        ],
        library_dirs=[*extra_library_dirs],
        libraries=['qpdf'],
        cxx_std=17,
    ),
)

if not cflags_defined:
    if sys.platform == 'cygwin':
        # On cygwin, use gnu++17 instead of c++17
        eca = extmodule.extra_compile_args
        eca[eca.index('-std=c++17')] = '-std=gnu++17'

    # Debug build
    # module[0].extra_compile_args.append('-g3')

    if qpdf_source_tree:
        for lib in extra_library_dirs:
            extmodule.extra_link_args.append(f'-Wl,-rpath,{lib}')  # type: ignore

if __name__ == '__main__':
    with ParallelCompile("PIKEPDF_NUM_BUILD_JOBS"):  # optional envvar
        setup(
            ext_modules=[extmodule],
            cmdclass={"build_ext": build_ext},  # type: ignore
        )



# SPDX-FileCopyrightText: 2022 James R. Barlow
# SPDX-License-Identifier: MPL-2.0

from __future__ import annotations


class DependencyError(Exception):
    "A third party dependency is needed to extract streams of this type."
